package org.bitbucket.bradleysmithllc.etlunit.feature.database;

import com.github.fge.jackson.JsonLoader;
import org.apache.commons.io.FileUtils;
import org.bitbucket.bradleysmithllc.etlunit.*;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.feature.Feature;
import org.bitbucket.bradleysmithllc.etlunit.feature.RuntimeOption;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.db.*;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database._assert.AssertHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database._assert.AssertRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.execute.execute_sql.ExecuteSqlHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.execute.execute_sql.ExecuteSqlRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.execute.execute_sql_procedure.ExecuteSqlProcedureHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.execute.execute_sql_procedure.ExecuteSqlProcedureRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.execute.execute_sql_script.ExecuteSqlScriptHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.execute.execute_sql_script.ExecuteSqlScriptRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.extract.ExtractHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.extract.ExtractRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.migrate.MigrateHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.migrate.MigrateRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.stage.StageHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.stage.StageRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.truncate.TruncateHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.json.database.truncate.TruncateRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.util.DimensionalHashMap;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.util.DimensionalMap;
import org.bitbucket.bradleysmithllc.etlunit.feature.file.FileRuntimeSupport;
import org.bitbucket.bradleysmithllc.etlunit.io.FileBuilder;
import org.bitbucket.bradleysmithllc.etlunit.io.file.*;
import org.bitbucket.bradleysmithllc.etlunit.metadata.MetaDataArtifact;
import org.bitbucket.bradleysmithllc.etlunit.metadata.MetaDataContext;
import org.bitbucket.bradleysmithllc.etlunit.metadata.MetaDataPackageContext;
import org.bitbucket.bradleysmithllc.etlunit.util.StringUtils;
import org.bitbucket.bradleysmithllc.json.validator.ResourceNotFoundException;
import org.bitbucket.bradleysmithllc.etlunit.parser.*;
import org.bitbucket.bradleysmithllc.etlunit.util.IOUtils;

import javax.inject.Inject;
import javax.inject.Named;
import java.io.File;
import java.io.IOException;
import java.math.BigDecimal;
import java.sql.*;
import java.sql.Date;
import java.util.*;

public class DatabaseClassListener extends NullClassListener implements ExtractHandler, StageHandler, TruncateHandler, MigrateHandler, AssertHandler, ExecuteSqlHandler, ExecuteSqlProcedureHandler, ExecuteSqlScriptHandler {
	public static final String DEFAULT_MODE = null;
	public static final String LAST_EXTRACT_FILE_SCHEMA = "lastExtractFileSchema";
	public static final String LAST_EXTRACT_TABLE = "LAST_EXTRACT_TABLE";

	private final DatabaseFeatureModule parent;
	private final DatabaseConfiguration databaseConfiguration;
	private DatabaseRuntimeSupport databaseRuntimeSupport;
	private RuntimeSupport runtimeSupport;
	private FileRuntimeSupport fileRuntimeSupport;
	protected Log applicationLog;
	private JDBCClient jdbcClient;

	private boolean prepareDatabase = true;
	private boolean refreshStageData = false;
	private boolean forceInitializeDatabase = false;

	private DataFileManager dataFileManager;

	private final AssertHandler assertHandler;
	private final DbOperationsExecutor executeHandler;

	public DatabaseClassListener(DatabaseFeatureModule parent, DatabaseConfiguration databaseConfiguration, AssertHandler object1, DbOperationsExecutor object) {
		assertHandler = object1;
		executeHandler = object;
		this.parent = parent;
		this.databaseConfiguration = databaseConfiguration;
	}

	@Inject
	public void receiveDataFileManager(DataFileManager manager) {
		dataFileManager = manager;
	}

	@Inject
	public void receiveRuntimeSupport(RuntimeSupport manager) {
		runtimeSupport = manager;
	}

	@Inject
	public void receiveFileRuntimeSupport(FileRuntimeSupport manager) {
		fileRuntimeSupport = manager;
	}

	@Inject
	public void setExportStageData(@Named("database.refreshStageData") RuntimeOption option) {
		refreshStageData = option.isEnabled();

		applicationLog.info(refreshStageData ? "Refreshing stage data" : "Not refreshing stage data");
	}

	@Inject
	public void setPrepareDatabase(@Named("database.prepareDatabase") RuntimeOption option) {
		prepareDatabase = option.isEnabled();

		applicationLog.info(prepareDatabase ? "Preparing database" : "Not preparing database");
	}

	@Inject
	public void setForceInitializeDatabase(@Named("database.forceInitializeDatabase") RuntimeOption option) {
		forceInitializeDatabase = option.isEnabled();

		applicationLog.info(forceInitializeDatabase ? "Forcing database" : "Not forcing database");
	}

	@Inject
	public void setJdbcClient(JDBCClient client) {
		jdbcClient = client;
	}

	@Inject
	public void setApplicationLog(@Named("applicationLog") Log log) {
		applicationLog = log;
	}

	@Inject
	public void setDatabaseRuntimeSupport(DatabaseRuntimeSupport databaseRuntimeSupport) {
		this.databaseRuntimeSupport = databaseRuntimeSupport;
	}

	@Override
	public action_code process(ETLTestOperation etlTestOperation, ETLTestValueObject etlTestValueObject, VariableContext context, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		return action_code.defer;
	}

	@Override
	public void begin(final ETLTestMethod mt, VariableContext context) throws TestAssertionFailure, TestExecutionError, TestWarning {
		// if this method, or class is annotated with @NoDatabase, then no further work is done
		if (mt.getAnnotations("@NoDatabase").size() > 0 || mt.getTestClass().getAnnotations("@NoDatabase").size() > 0) {
			// do nothing
			return;
		}

		// create a view of the annotations which has them in priority order descending
		List<ETLTestAnnotation> annoList = new ArrayList<ETLTestAnnotation>(mt.getTestClass().getAnnotations("@Database"));
		annoList.addAll(mt.getAnnotations("@Database"));

		// before starting a method, initialize any database support necessary.
		// get a list of databases to prepare
		DimensionalMap<DatabaseConnection, String, Object> dbMap = new DimensionalHashMap<DatabaseConnection, String, Object>();

		if (databaseConfiguration == null) {
			throw new TestExecutionError("Database feature lacks configuration", DatabaseConstants.ERR_MISSING_DATABASE_CONFIGURATION);
		}

		Iterator<ETLTestAnnotation> it = annoList.iterator();

		while (it.hasNext()) {
			ETLTestAnnotation anno = it.next();

			ETLTestValueObject annoet = anno.getValue();
			String id = annoet.query("id").getValueAsString();

			ETLTestValueObject modes = annoet.query("modes");

			DatabaseConnection dc = databaseConfiguration.getDatabaseConnection(id);

			if (dc == null) {
				throw new TestExecutionError("Undeclared connection id " + id);
			}

			// clear previous modes for this connection
			dbMap.remove(dc);

			if (modes != null) {
				List<String> list = modes.getValueAsListOfStrings();

				for (String mode : list) {
					dbMap.put(dc, mode, "");
				}
			} else {
				dbMap.put(dc, DEFAULT_MODE, "");
			}
		}

		if (dbMap.size() == 0) {
			// nothing to do at all
			return;
		}

		// prepare the databases
		Iterator<DatabaseConnection> set = dbMap.keySet().iterator();

		while (set.hasNext()) {
			DatabaseConnection dc = set.next();

			// look up the implementation and request a prepare
			DatabaseImplementation impl = parent.getImplementation(dc.getId());

			if (impl == null) {
				throw new TestExecutionError("Connection id " + dc.getId() + " does not have a valid implementation: " + dc.getImplementationId(), DatabaseConstants.ERR_MISSING_IMPLEMENTATION);
			}

			Map<String, Object> lmap = dbMap.get(dc);

			for (final String mode : lmap.keySet()) {
				// validate the modes
				// check that the configuration allows for any mode (indicated by no mode specified on the connection)
				final DatabaseConnection databaseConnection = databaseConfiguration.getDatabaseConnection(dc.getId());

				// potentially initialize
				if (prepareDatabase) {
					if (!parent.initializedDbs.containsKey(dc.getId(), mode)) {
						applicationLog.info("Initializing database: " + dc.getId() + (mode == null ? "" : ("." + mode)));

						if (forceInitializeDatabase || parent.databaseOutOfDate(mode, databaseConnection) || impl.getDatabaseState(databaseConnection, mode) == DatabaseImplementation.database_state.fail) {
							DenyAllOperationsOperationRequest request = new DenyAllOperationsOperationRequest() {
								@Override
								public DatabaseImplementation.InitializeRequest getInitializeRequest() throws UnsupportedOperationException {
									return new DatabaseImplementation.InitializeRequest() {
										public String getMode() {
											return mode;
										}

										public DatabaseConnection getConnection() {
											return databaseConnection;
										}

										public ETLTestMethod getTestMethod() {
											return mt;
										}

										public File resolveSchemaReference(String ddlRef) {
											throw new UnsupportedOperationException();
										}
									};
								}
							};

							// we have to issue createDatabase, createSchema, complete, dropConstraints and materializeViews
							impl.processOperation(DatabaseImplementation.operation.createDatabase, request);
							impl.processOperation(DatabaseImplementation.operation.preCreateSchema, request);

							// create the schema
							jdbcClient.useStatement(databaseConnection, mode, new JDBCClient.StatementClient() {
								public void connection(Connection conn, Statement st, DatabaseConnection connection, String mode, int id) throws Exception {
									for (String ref : databaseConnection.getSchemaScripts()) {
										applicationLog.info("Processing schema source: " + ref);

										DDLSourceRef sourceRef = new DDLSourceRef(databaseConnection.getImplementationId(), databaseConnection.getId(), ref);

										SQLAggregator SQLAggregator = parent.resolveDDLRef(sourceRef, databaseConnection);

										int count = 0;

										SQLAggregator.Aggregator agg = SQLAggregator.getStatementAggregator();

										while (agg.hasNext()) {
											count++;

											SQLAggregator.FileRef fRef = agg.next();

											String sql = fRef.getLine();

											try {
												st.execute(sql);
											} catch (Throwable thr) {
												throw new TestExecutionError("Error processing DDL reference: " + fRef.describe(), DatabaseConstants.ERR_DDL_APPLICATION_FAILURE, thr);
											}
										}

										applicationLog.info("Processed [" + count + "] statements");
									}
								}
							});

							impl.processOperation(DatabaseImplementation.operation.postCreateSchema, request);
							impl.processOperation(DatabaseImplementation.operation.completeDatabaseInitialization, request);

							impl.processOperation(DatabaseImplementation.operation.materializeViews, request);

							try {
								// this is here to force the database meta data to be refreshed after schema creation
								// otherwise the process operation calls after this one will not have a database loaded
								// in the connection.  This call may not be cached since the schema was just created
								getTableMetaData2(databaseConnection, mode, impl, false);
							} catch (Exception e) {
								throw new TestExecutionError("Failure reading database meta data [" + databaseConnection.getId() + "]", DatabaseConstants.ERR_PROCESS_DATABASE_META_DATA, e);
							}

							impl.processOperation(DatabaseImplementation.operation.dropConstraints, request);
							parent.updateDatabaseFlag(mode, databaseConnection);
						}

						parent.initializedDbs.put(dc.getId(), mode, "");
					}

					// prepare the database for test
					applicationLog.info("Purging database for test: " + dc.getId() + (mode == null ? "" : ("." + mode)));

					DenyAllOperationsOperationRequest denyAllOperationsOperationRequest = new DenyAllOperationsOperationRequest() {
						@Override
						public DatabaseImplementation.DatabaseRequest getDatabaseRequest() throws UnsupportedOperationException {
							return new DatabaseImplementation.DatabaseRequest() {
								public ETLTestMethod getTestMethod() {
									return mt;
								}

								public String getMode() {
									return mode;
								}

								public DatabaseConnection getConnection() {
									return databaseConnection;
								}
							};
						}
					};

					impl.processOperation(DatabaseImplementation.operation.preCleanTables, denyAllOperationsOperationRequest);

					/* This lookup for database may be cached since tis execution path may not have come through the initialize step */
					Database database = null;
					try {
						database = getTableMetaData2(databaseConnection, mode, impl, true);

					} catch (Exception e) {
						throw new TestExecutionError("Failure reading database meta data [" + databaseConnection.getId() + "]", DatabaseConstants.ERR_PROCESS_DATABASE_META_DATA, e);
					}

					int tblCounter = 0;

					for (Catalog catalog : database.getCatalogs()) {
						for (Schema schema : catalog.getSchemas()) {
							for (Table table : schema.getTables()) {
								try {
									if (table.getType() != Table.type.sql) {
										tblCounter++;
										impl.purgeTableForTest(databaseConnection, mode, table);
									}
								} catch (Exception e) {
									throw new TestExecutionError("Failure preparing table [" + table.toString() + "] for test", DatabaseConstants.ERR_PREPARE_TABLES, e);
								}
							}
						}
					}

					applicationLog.info("Purged [" + tblCounter + "] tables");
					impl.processOperation(DatabaseImplementation.operation.postCleanTables, denyAllOperationsOperationRequest);
				} else if (parent.databaseOutOfDate(mode, databaseConnection)) {
					throw new TestExecutionError("Prepare database option is false, but the database schema has changed.", DatabaseConstants.ERR_INVALID_SCHEMA_STATE);
				}
			}

			/* Ping this once more since prepareDatabase might be turned off which would cause the entire block above not to be processed */
			if (!dc.hasDatabase()) {
				try {
					getTableMetaData2(dc, null, impl, true);
				} catch (Exception e) {
					throw new TestExecutionError("Failure reading database meta data [" + dc.getId() + "]", DatabaseConstants.ERR_PROCESS_DATABASE_META_DATA, e);
				}
			}
		}

		/* store database information in the variable context for use by other features
		 * These should look like this to the clients:
		 * database: {
		 * 	connections: {
		 * 		connectionName: {
		 * 			target: databaseConnection
		 * 		}
		 *	 }
		 *	}
		 *	Where the database connection is a facade which only responds to the configured
		 *	mode (target in the example).  The 'modeless' connection uses the
		 *	reserved name 'default'
		 */

		// create the database container
		ETLTestValueObjectBuilder builder = new ETLTestValueObjectBuilder();
		builder = builder.object().key("connections").object();

		// iterate through all required connections and modes
		for (DatabaseConnection key : dbMap.keySet()) {
			Map<String, Object> lmap = dbMap.get(key);

			// create the base connection object
			// this will contain modes
			builder = builder.key(key.getId()).object();
			// run through all modes, checking for the special null mode which gets converted
			// to 'default'

			for (String mode : lmap.keySet()) {
				if (mode == null) {
					builder.key("default").object();
				} else {
					builder.key(mode).object();
				}

				// add properties
				parent.propogateProperties(builder, key, mode);
				builder.endObject();
			}

			builder.endObject();
		}

		builder.endObject().endObject();

		context.declareAndSetValue("database", builder.toObject());
	}

	enum source_mode {table, sql}

	public action_code extract(ExtractRequest operation, final ETLTestMethod mt, final ETLTestOperation op, final ETLTestValueObject obj, final VariableContext context, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		final String source;
		source_mode smode = source_mode.table;

		String sqlScript = operation.getSqlScript();
		String extractOperationSql = operation.getSql();

		if (sqlScript != null) {
			// load the sql into the source variable
			try {
				File sqlFile = databaseRuntimeSupport.getSourceScriptForCurrentTest(sqlScript);

				if (!sqlFile.exists()) {
					throw new TestExecutionError("Sql script not found: " + sqlFile, DatabaseConstants.ERR_MISSING_SQL_SCRIPT);
				}

				MetaDataContext dataMetaContext = parent.getDatabaseMetaContext();

				MetaDataPackageContext dataPackageContext = dataMetaContext.createPackageContextForCurrentTest(MetaDataPackageContext.path_type.test_source);
				MetaDataArtifact art;

				try {
					art = dataPackageContext.createArtifact("sql", databaseRuntimeSupport.getSourceDataDirectoryForCurrentTest());
					art.createContent(sqlFile.getName()).referencedByCurrentTest("sql-extract-script");

				} catch (IOException e) {
					throw new TestExecutionError("Error getting data context", e);
				}

				source = IOUtils.readFileToString(sqlFile);
				smode = source_mode.sql;
			} catch (IOException e) {
				throw new TestExecutionError("", e);
			}
		} else if (extractOperationSql != null) {
			smode = source_mode.sql;
			source = extractOperationSql;
		} else {
			source = operation.getSourceTable();
		}

		//final String sourceTable = source;
		final String target = operation.getTarget();
		final source_mode sourcemode = smode;

		ExtractRequest.ColumnListMode lm = operation.getColumnListMode();

		final List<String> list = new ArrayList(operation.getColumnList());

		if (list.size() != 0 && lm == null) {
			// exclude is the default
			lm = ExtractRequest.ColumnListMode.EXCLUDE;
		} else if (lm != null && list.size() == 0) {
			throw new TestExecutionError("Extract operation does not allow column-list-mode without column-list");
		}

		final ExtractRequest.ColumnListMode listMode = lm;

		/**
		 * This isn't an issue anymore.  We will simply extract the full and create a subview for sql and script sources.
		 */
		//if ((listMode != null || list.size() != 0) && sourcemode == source_mode.sql) {
		//throw new TestExecutionError("Extract operation does not allow column-list-mode using sql");
		//}

		// look for optional data
		final String targetFile = operation.getTargetFile();

		ConnectionMode connMode = getConnection(parent, obj, context, op.getTestMethod()).get(0);

		final DatabaseConnection conn = connMode.conn;

		final String mode = connMode.modes.get(0);

		// now send it off to the implementation
		final DatabaseImplementation impl = getDatabaseImplementation(parent, conn);

		String aschema = operation.getSourceSchema();

		boolean restrictedSchema = impl.restrictToOwnedSchema(conn);
		if (aschema == null) {
			aschema = restrictedSchema ? null : impl.getDefaultSchema(conn, mode);
		}

		final String schema = aschema;

		final File targetBase;

		if (targetFile != null) {
			File f = new File(targetFile);

			if (f.isAbsolute()) {
				targetBase = f;
			} else {
				targetBase = databaseRuntimeSupport.getGeneratedDataSet(conn, connMode.getPrettyMode(), targetFile);
			}
		} else {
			targetBase = null;
		}

		// get the relational table definition
		try {
			Database meta = conn.getDatabase();

			Catalog catalog = meta.getCatalog(null);

			final String key;

			Table ltable;

			String sqlId = null;

			if (sourcemode == source_mode.sql) {
				sqlId = sqlScript;

				if (extractOperationSql != null) {
					// in this case, use the identity of the operation to uniquely identify the meta format
					sqlId = op.getQualifiedName();
				}

				sqlId = sqlId.replace(".", "-");

				key = (conn.getId() + "__SQL__" + sqlId).toUpperCase();
				Catalog sqlCatalog = catalog.getDatabase().getCatalog("__SQL__");
				Schema catalogSchema = sqlCatalog.getSchema(null);
				ltable = catalogSchema.getTable(sqlId);

				if (ltable == null) {
					TableImpl tbl = new TableImpl(sqlCatalog, catalogSchema, sqlId, Table.type.sql);
					tbl.setSqlScriptSource(source);
					catalogSchema.addTable(tbl);
					ltable = tbl;
				} else {
					// always update the sql for the source query so if the script is updated in the project that will be recorded
					ltable.setSqlScriptSource(source);
				}
			} else {
				key = (conn.getId() + "." + (aschema != null ? (aschema.toUpperCase() + ".") : "") + source).toUpperCase();

				Table table1 = catalog.getSchema(restrictedSchema ? "USER" : aschema).getTable(source);

				if (!restrictedSchema && aschema == null && table1 == null) {
					// try to find the table without a schema
					table1 = getSchemaLessTableData(conn, mode, impl, meta, source);
				}

				ltable = table1;
			}

			// validate meta data exists
			if (ltable == null) {
				throw new TestExecutionError("Database does not export table " + key + " in it's meta data", DatabaseConstants.ERR_MISSING_SCHEMA_TABLE);
			}

			loadTable(ltable, mode, impl);

			final Table table = ltable;

			final File dataTarget;

			if (targetFile == null) {
				dataTarget = databaseRuntimeSupport.getSourceDataSetForCurrentTest(target);
			} else {
				dataTarget = targetBase;
			}

			applicationLog.info("Extracting source " + (sourcemode == source_mode.table ? "table" : "query") + "[" + key + "]" + " to local dataset [" + dataTarget.getAbsolutePath() + "]");

			// export the data
			jdbcClient.useStatement(conn, mode, new JDBCClient.StatementClient() {
				public void connection(Connection conn, Statement st, DatabaseConnection connection, String mode, int id) throws Exception {
					DataFileSchema fileSchema = loadSchema(table, source, obj);

					applicationLog.info("Resolved data file schema: " + fileSchema.getId());

					fileSchema.setColumnDelimiter(databaseConfiguration.getDataFileColumnDelimiter());
					fileSchema.setRowDelimiter(databaseConfiguration.getDataFileRowDelimiter());
					fileSchema.setNullToken(databaseConfiguration.getDataFileNullToken());

					Table useTable = table;

					/**
					 * The subview for sql queries must happen after extraction - not before.  I.E., we can't modify the source query only the results.
					 */
					if (listMode != null && table.getType() != Table.type.sql) {
						useTable = table.createSubView(list, listMode);
						switch (listMode) {
							case INCLUDE:
								fileSchema = fileSchema.createSubViewIncludingColumns(list, fileSchema.getId() + "-subIncluding-" + op.getQualifiedName());
								break;
							case EXCLUDE:
								fileSchema = fileSchema.createSubViewExcludingColumns(list, fileSchema.getId() + "subExcluding" + op.getQualifiedName());
								break;
						}

						// store the generated file schema for reference
						fileRuntimeSupport.persistGeneratedDataFileSchemaForCurrentTest(fileSchema);
					}

					// store the schema regardless of whether we had a sql assertion or not
					context.declareAndSetValue(LAST_EXTRACT_FILE_SCHEMA, new ETLTestValueObjectImpl(fileSchema));
					context.declareAndSetValue(LAST_EXTRACT_TABLE, new ETLTestValueObjectImpl(table));

					DataFile ff = dataFileManager.loadDataFile(dataTarget, fileSchema);

					String variableName = "extract.columns." + target;
					applicationLog.info("Setting context variable for " + variableName);
					context.declareAndSetValue(variableName, new ETLTestValueObjectImpl(useTable.getTableColumnNames()));

					String sqlSelect = useTable.createSQLSelect(impl);

					if (sqlSelect == null) {
						throw new TestExecutionError("Null sql select created by table: " + useTable.getQualifiedName());
					}

					applicationLog.info("Using sql select for object(" + key + " in [" + connection.getId() + "." + mode + "]):\n" + sqlSelect);

					long startTime = System.currentTimeMillis();
					int rowsRead = 0;

					DataFileWriter ffwriter = ff.getWriter(new JdbcTypeConverter());

					try {
						ResultSet rs = st.executeQuery(sqlSelect);

						ResultSetMetaData rsmd = rs.getMetaData();

						int cols = rsmd.getColumnCount();

						try {
							Map<String, Object> rowData = new HashMap<String, Object>();

							while (rs.next()) {
								rowsRead++;

								for (int i = 1; i <= cols; i++) {
									String colName = rsmd.getColumnName(i);
									Object data = rs.getObject(i);

									rowData.put(colName, data);
								}

								ffwriter.addRow(rowData);
							}
						} finally {
							rs.close();
						}
					} finally {
						ffwriter.close();
					}

					// in the case of a sql query, we must create a view of the file to pass on
					if (table.getType() == Table.type.sql && listMode != null) {
						applicationLog.info("Copying sql extract to match column list");

						DataFileSchema targetFileSchema = fileSchema;

						switch (listMode) {
							case INCLUDE:
								targetFileSchema = fileSchema.createSubViewIncludingColumns(list, fileSchema.getId() + "-sql-subIncluding-" + op.getQualifiedName());
								break;
							case EXCLUDE:
								targetFileSchema = fileSchema.createSubViewExcludingColumns(list, fileSchema.getId() + "-sql-subExcluding-" + op.getQualifiedName());
								break;
						}

						fileRuntimeSupport.persistGeneratedDataFileSchemaForCurrentTest(fileSchema);

						context.setValue(LAST_EXTRACT_FILE_SCHEMA, new ETLTestValueObjectImpl(targetFileSchema));

						// create a temporary data file
						File ddsql = databaseRuntimeSupport.getGeneratedDataSet(connection, mode, target + ".sqldat");

						// copy the output to the generated temp
						FileUtils.copyFile(dataTarget, ddsql);

						// create a data file off the temp sql output with the full schema
						DataFile ffSql = dataFileManager.loadDataFile(ddsql, fileSchema);

						// create a data file pointing to the target with the final column list
						DataFile ffExt = dataFileManager.loadDataFile(dataTarget, targetFileSchema);

						dataFileManager.copyDataFile(ffSql, ffExt);
					}

					long endTime = System.currentTimeMillis();
					applicationLog.debug("Selected [" + rowsRead + "] rows in [" + (endTime - startTime) + "] ms.");
				}
			});
		} catch (TestExecutionError e) {
			throw e;
		} catch (Exception e) {
			throw new TestExecutionError("", e);
		}

		return action_code.handled;
	}

	private DataFileSchema loadSchema(Table table, String source, ETLTestValueObject obj) throws TestExecutionError {
		// look for the meta info in the project
		try {
			// look it up in the project
			return databaseRuntimeSupport.getRelationalMetaInfo(table.getCatalog().getDatabase().getDatabaseConnection(), table.getSchema().getName(), table.getName(), source, obj);
		} catch (ResourceNotFoundException exc) {
			// use the relational definition
			return table.getDataFileSchema();
		} catch (IllegalArgumentException exc)
		{
			throw new TestExecutionError("", DatabaseConstants.ERR_REFERENCE_FILE_NOT_FOUND);
		}
	}

	private void loadTable(final Table table, String mode, final DatabaseImplementation impl) throws Exception {

		if (table.getTableColumns().size() == 0) {
			applicationLog.info("Updating database meta data for target [" + table.getCatalog().getDatabase().getDatabaseConnection().getId() + "." + mode + "." + table.getSchema().getName() + "." + table.getName() + "]");

			jdbcClient.useStatement(table.getCatalog().getDatabase().getDatabaseConnection(), mode, new JDBCClient.StatementClient() {
				public void connection(Connection conn, Statement st, DatabaseConnection connection, String mode, int id) throws Exception {
					if (table.getType() != Table.type.sql) {
						DatabaseMetaData meta = conn.getMetaData();

						boolean restricted = impl.restrictToOwnedSchema(connection);

						// issue a select for the table and check the meta data for column info
						String sql = "SELECT * FROM " + impl.escapeQualifiedIdentifier(table) + " WHERE 1 = 2";
						applicationLog.info("Using meta query string: " + sql);
						ResultSet rs = st.executeQuery(sql);

						try {
							ResultSetMetaData rsmd = rs.getMetaData();

							for (int col = 1; col <= rsmd.getColumnCount(); col++) {
								String columnName = rsmd.getColumnName(col);
								int columnType = rsmd.getColumnType(col);
								int nullable = rsmd.isNullable(col);
								boolean autoIncrement = rsmd.isAutoIncrement(col);

								TableColumn tc = new TableColumnImpl(
										table,
										columnName,
										columnType,
										nullable == ResultSetMetaData.columnNullable,
										autoIncrement
								);

								table.addColumn(tc);
							}
						} finally {
							rs.close();
						}

						// add all primary key columns (only the first is of interest)
						rs = meta.getPrimaryKeys(null, table.getSchema().getName(), table.getName());

						try {
							String firstPk = null;

							while (rs.next()) {
								String columnName = rs.getString(4);
								String pkName = rs.getString(6);

								if (firstPk == null) {
									firstPk = pkName;
								} else if (!firstPk.equals(pkName)) {
									// we don't care about the rest
									break;
								}

								table.addPrimaryKeyColumn(table.getColumn(columnName));
							}
						} finally {
							rs.close();
						}
					} else {
						// in the case of a SQL query, execute the query once and analyze the result set meta data
						String sqlSelect = table.createSQLSelect(impl);
						ResultSet rs = st.executeQuery(sqlSelect);

						ResultSetMetaData rsmd = rs.getMetaData();

						try {
							for (int i = 1; i <= rsmd.getColumnCount(); i++) {
								String columnName = rsmd.getColumnLabel(i);
								int columnType = rsmd.getColumnType(i);

								// for the sql data set, there is no primary key
								TableColumn tc = new TableColumnImpl(
										table,
										columnName,
										columnType,
										true,
										false
								);

								table.addColumn(tc);
							}
						} finally {
							rs.close();
						}
					}
				}
			});
		}

		// look for the meta info in the project
		try {
			DataFileSchema ffs = databaseRuntimeSupport.getRelationalMetaInfo(table.getCatalog().getDatabase().getDatabaseConnection(), table.getSchema().getName(), table.getName(), null, null);

			applicationLog.info("Reading meta data from reference file: " + ffs.getId());

			// load the file and override the order by columns

			table.resetOrderColumns();
			for (String name : ffs.getOrderColumnNames()) {
				table.addOrderKeyColumn(table.getColumn(name));
			}

			table.resetPrimaryKeyColumns();
			for (String name : ffs.getKeyColumnNames()) {
				table.addPrimaryKeyColumn(table.getColumn(name));
			}

			table.setDataFileSchema(ffs);
		} catch (ResourceNotFoundException exc) {
			DataFileSchema ffs = databaseRuntimeSupport.createRelationalMetaInfo(table.getCatalog().getDatabase().getDatabaseConnection(), table.getSchema().getName(), table.getName());

			applicationLog.info("Could not find reference file - creating boilerplate: " + ffs.getId());

			// generate a source definition for this set

			// write out a json object which defines the connection id and table, and the columns
			ffs.setFormatType(DataFileSchema.format_type.delimited);

			for (TableColumn col : table.getTableColumns()) {
				DataFileSchema.Column sc = ffs.createColumn(col.getName());
				sc.setType(JdbcTypeConverter.getTypeName(col.getType()));

				switch (col.getType()) {
					case Types.BIGINT:
					case Types.INTEGER:
					case Types.SMALLINT:
					case Types.TINYINT:
					case Types.BIT:
						sc.setBasicType(DataFileSchema.Column.basic_type.integer);
						break;
					case Types.DECIMAL:
					case Types.DOUBLE:
					case Types.FLOAT:
					case Types.NUMERIC:
						sc.setBasicType(DataFileSchema.Column.basic_type.numeric);
						break;
				}

				ffs.addColumn(sc);
			}

			for (TableColumn colName : table.getPrimaryKeyColumns()) {
				ffs.addKeyColumn(colName.getName());
			}

			for (TableColumn colName : table.getOrderKeyColumns()) {
				ffs.addKeyColumn(colName.getName());
			}

			table.setDataFileSchema(ffs);

			// load into Jackson and pretty-print
			databaseRuntimeSupport.persistDataFileSchema(ffs);
		}
	}

	private Database getTableMetaData2(
			DatabaseConnection conn,
			String mode,
			final DatabaseImplementation impl,
			boolean useCached
	) throws Exception {
		if (conn.hasDatabase()) {
			return conn.getDatabase();
		}

		final Database database = new DatabaseImpl(conn, conn.getId());

		conn.setDatabase(database);

		MetaDataContext databaseMetaContext = parent.getDatabaseMetaContext();
		final MetaDataPackageContext packContext = databaseMetaContext.createPackageContext(conn.getId() + "." + (mode != null ? (mode) : "[default]"), MetaDataPackageContext.path_type.external_source);

		File databaseMetadataTarget = new FileBuilder(databaseRuntimeSupport.getGeneratedSourceDirectory(conn, null)).name(conn.getId() + "_metadata.json").file();
		final MetaDataArtifact art = packContext.createArtifact("schema", databaseMetadataTarget);

		if (!useCached) {
			applicationLog.info("Ignoring cached meta data for connection [" + conn.getId() + "]");
		}

		if (!useCached && databaseMetadataTarget.exists()) {
			applicationLog.info("Ignoring cached meta data for connection even though cache exists [" + conn.getId() + "]");
		} else if (useCached && !databaseMetadataTarget.exists()) {
			applicationLog.info("Cannot use cached meta data for connection [" + conn.getId() + "] - cache file does not exist [" + databaseMetadataTarget.getAbsolutePath() + "]");
		}

		if (useCached && databaseMetadataTarget.exists()) {
			applicationLog.info("Using cached meta data for connection [" + conn.getId() + "]");
			database.fromJson(JsonLoader.fromFile(databaseMetadataTarget));
		} else {
			applicationLog.info("Querying meta data for connection [" + conn.getId() + "]");

			jdbcClient.useConnection(conn, mode, new JDBCClient.ConnectionClient() {
				public void connection(Connection conn, DatabaseConnection connection, String mode, int id) throws Exception {
					DatabaseMetaData meta = conn.getMetaData();

					boolean restrictToOwnedSchema = impl.restrictToOwnedSchema(connection);
					String schemaPattern = restrictToOwnedSchema ? StringUtils.getSafeNullableUppercase(impl.getLoginName(connection, mode)) : null;
					ResultSet rs = meta.getTables(null, schemaPattern, null, null);

					try {
						while (rs.next()) {
							//String catalog = rs.getString(1);  // I don't know what this affects yet.
							String dbTableSchema = rs.getString(2);
							String tableName = rs.getString(3);
							String tableType = rs.getString(4);

							String tableSchema = restrictToOwnedSchema ? "USER" : dbTableSchema;

							if (tableType != null) {
								String tableUpper = tableType.toUpperCase();

								Table.type type = impl.translateTableType(tableUpper);

								if (type != null) {
									Catalog databaseCatalog = database.getCatalog(null);
									Schema sch = databaseCatalog.getSchema(tableSchema);
									sch.setVirtual(restrictToOwnedSchema);

									TableImpl tbl = new TableImpl(databaseCatalog, sch, tableName, type);

									if (impl.isTableTestVisible(connection, mode, tbl)) {
										sch.addTable(tbl);

										applicationLog.debug("Driver classified table " + impl.escapeQualifiedIdentifier(tbl) + " as [" + type + "]");
										// add to meta data

										art.createContent(tbl.getQualifiedName());
									} else {
										applicationLog.debug("Driver excluded table: " + impl.escapeQualifiedIdentifier(tbl));
									}
								}
							}
						}
					} finally {
						rs.close();
					}
				}
			});
		}

		return database;
	}

	private Table getSchemaLessTableData(DatabaseConnection conn, String mode, final DatabaseImplementation impl, final Database relationalSet, final String tableName) throws Exception {
		Table table = relationalSet.getCatalog(null).getSchema(null).getTable(tableName);

		if (table != null) {
			return table;
		}

		jdbcClient.useConnection(conn, mode, new JDBCClient.ConnectionClient() {
			public void connection(Connection conn, DatabaseConnection connection, String mode, int id) throws Exception {
				DatabaseMetaData meta = conn.getMetaData();

				ResultSet rs = meta.getTables(null, null, tableName, null);

				try {
					if (rs.next()) {
						//String tableCatalog = rs.getString(1); // Ditto
						String tableSchema = rs.getString(2);
						String tableName = rs.getString(3);
						String tableType = rs.getString(4);

						if (tableType != null) {
							String tableUpper = tableType.toUpperCase();

							Table.type type = impl.translateTableType(tableUpper);

							if (type != null) {
								Catalog databaseCatalog = relationalSet.getCatalog(null);
								Schema sch = databaseCatalog.getSchema(tableSchema);

								TableImpl tbl = (TableImpl) sch.getTable(tableName);

								if (tbl == null) {
									tbl = new TableImpl(databaseCatalog, sch, tableName, type);

									((SchemaImpl) sch).addTable(tbl);

									String key1 = (connection.getId() + "." + tableSchema + '.' + tableName).toUpperCase();

									applicationLog.debug(key1);
								}

								relationalSet.getCatalog(null).getSchema(null).addTable(tbl);
							}
						}
					} else {
						throw new TestExecutionError("Could not locate table without schema: " + tableName, DatabaseConstants.ERR_SCHEMALESS_TABLE_NOT_FOUND);
					}

					if (rs.next()) {
						throw new TestExecutionError("Table reference ambiguous without schema: " + tableName, DatabaseConstants.ERR_AMBIGUOUS_TABLE);
					}
				} finally {
					rs.close();
				}
			}
		});


		return relationalSet.getCatalog(null).getSchema(null).getTable(tableName);
	}

	private void stage(ETLTestMethod mt, final ETLTestOperation etlTestOperation, ETLTestValueObject etlTestValueObject1, StageRequest stageOperation, VariableContext context, ExecutionContext econtext) throws TestExecutionError, TestAssertionFailure, TestWarning {
		MetaDataContext dataMetaContext = parent.getDataMetaContext();

		MetaDataPackageContext dataPackageContext = dataMetaContext.createPackageContextForCurrentTest(MetaDataPackageContext.path_type.test_source);
		MetaDataArtifact art;

		try {
			art = dataPackageContext.createArtifact("data", databaseRuntimeSupport.getSourceDataDirectoryForCurrentTest());
		} catch (IOException e) {
			throw new TestExecutionError("Error getting data context", e);
		}

		final String sourceFile = stageOperation.getSourceFile();
		String source = stageOperation.getSource();
		String targetTable = stageOperation.getTargetTable();

		// check the runtime option for refresh data.  If so, convert this call into an extract operation
		if (refreshStageData) {
			art.createContent(source).referencedByCurrentTest("assertion-refresh");
			ETLTestValueObject copy = etlTestValueObject1.copy();

			ETLTestValueObjectBuilder cbuilder = new ETLTestValueObjectBuilder(copy);

			//!!
			cbuilder = cbuilder.key("source-table").value(etlTestValueObject1.query("target-table"));
			//!!
			cbuilder.removeKey("target-table");

			if (etlTestValueObject1.query("target-schema") != null) {
				cbuilder = cbuilder.key("source-schema").value(etlTestValueObject1.query("target-schema"));
				cbuilder.removeKey("target-schema");
			}

			//!!
			cbuilder = cbuilder.key("target").value(etlTestValueObject1.query("source"));
			//!!
			cbuilder.removeKey("source");

			// check for multiple modes specified.  If present, replace with a single mode reference
			if (etlTestValueObject1.query("modes") != null) {
				cbuilder.removeKey("modes").key("mode").value(etlTestValueObject1.query("modes").getValueAsList().get(0));
			}

			copy = cbuilder.toObject();

			ETLTestOperation extract_op = etlTestOperation.createSibling("extract", copy);

			econtext.process(extract_op, context);

			return;
		}

		// look for connections and connection-id or {mode, modes}
		if (stageOperation.getConnections() != null &&
				(stageOperation.getMode() != null || stageOperation.getModes().size() != 0 || stageOperation.getConnectionId() != null)
				) {
			throw new TestExecutionError("Attribute connections may not be used in conjunction with connections, mode or modes");
		}

		File dataFile = null;

		if (sourceFile != null) {
			dataFile = new File(sourceFile);
		} else {
			dataFile = databaseRuntimeSupport.getSourceDataSet(etlTestOperation.getTestClass().getPackage(), source);
		}

		art.createContent(dataFile.getName()).referencedByCurrentTest("stage-source");

		// look for optional data
		List<ConnectionMode> connectionModes = getConnection(parent, etlTestValueObject1, context, etlTestOperation.getTestMethod());

		for (final ConnectionMode connectionMode : connectionModes) {
			// now send it off to the implementation
			final DatabaseImplementation impl = getDatabaseImplementation(parent, connectionMode.conn);

			// get the relational table definition
			List<String> modeList = connectionMode.modes;

			if (modeList == null) {
				modeList = new ArrayList<String>();
				modeList.add(null);
			}

			for (final String thisMode : modeList) {
				final String schema;

				String targetSchema = stageOperation.getTargetSchema();
				if (targetSchema != null) {
					schema = targetSchema;
				} else {
					schema = impl.restrictToOwnedSchema(connectionMode.conn) ? null : impl.getDefaultSchema(connectionMode.conn, thisMode);
				}

				final String key = (connectionMode.connectionId + "." + (schema != null ? (schema.toUpperCase() + ".") : "") + targetTable).toUpperCase();

				try {
					Database meta = connectionMode.conn.getDatabase();

					Table t1 = meta.getCatalog(null).getSchema(impl.restrictToOwnedSchema(connectionMode.conn) ? "USER" : schema).getTable(targetTable);

					if (schema == null && t1 == null) {
						// try to find the table without a schema
						t1 = getSchemaLessTableData(connectionMode.conn, thisMode, impl, meta, targetTable);
					}

					// load column and key data if not loaded already
					if (t1 == null) {
						throw new TestExecutionError("Database does not export table " + key + " in it's meta data", DatabaseConstants.ERR_MISSING_SCHEMA_TABLE);
					}

					// make sure it is populated
					loadTable(t1, thisMode, impl);

					final Table table = t1;

					applicationLog.info("Loading (" + key + ") data set (" + sourceFile + ")");

					// get the actual data file
					if (!dataFile.exists()) {
						throw new TestExecutionError(dataFile.getAbsolutePath(), DatabaseConstants.ERR_MISSSING_DATA_FILE);
					}

					DataFileSchema fileSchema = loadSchema(table, IOUtils.removeExtension(dataFile), etlTestValueObject1);

					final DataFile ff = dataFileManager.loadDataFile(dataFile, fileSchema);

					// open a jdbc connection and run away with it
					jdbcClient.usePreparedStatement(connectionMode.conn, thisMode, new JDBCClient.PreparedStatementClient() {
						public String prepareText() {
							String sqlInsert = table.createSQLInsert(impl);
							applicationLog.info("Using sql insert for object(" + key + ") in  [" + connectionMode.connectionId + "." + thisMode + "] - data set (" + sourceFile + "):\n" + sqlInsert);

							return sqlInsert;
						}

						public void connection(Connection conn, PreparedStatement pst, DatabaseConnection connection, String mode, int id) throws Exception {
							impl.prepareConnectionForInsert(conn, table, connection, mode);

							List<TableColumn> columns = table.getTableColumns();

							DataFile.FileData fdata = ff.getFileData();

							try {
								Iterator<DataFile.FileRow> dataIt = fdata.iterator();

								int rowNum = 0;
								int rowsInserted = 0;
								long startTime = System.currentTimeMillis();

								while (dataIt.hasNext()) {
									rowNum++;
									rowsInserted++;

									Map<String, String> rowData = dataIt.next().getData();

									for (int i = 0; i < columns.size(); i++) {
										TableColumn col = columns.get(i);

										int columnType = col.getType();

										String val = rowData.get(col.getName());

										// TMI!!
										//applicationLog.info("Processing value: " + val);

										int parameterIndex = i + 1;
										if (val == null) {
											// set this to null in the insert
											pst.setNull(parameterIndex, columnType);
										} else {
											try {
												switch (columnType) {
													case Types.INTEGER:
														pst.setInt(parameterIndex, Integer.parseInt(val));
														break;
													case Types.BIGINT:
														pst.setLong(parameterIndex, Long.parseLong(val));
														break;
													case Types.SMALLINT:
														pst.setShort(parameterIndex, Short.parseShort(val));
														break;
													case Types.TINYINT:
														pst.setByte(parameterIndex, Byte.parseByte(val));
														break;
													case Types.VARCHAR:
													case Types.CHAR:
													case Types.LONGNVARCHAR:
														pst.setString(parameterIndex, val);
														break;
													case Types.CLOB:
														pst.setString(parameterIndex, val);
														break;
													case Types.TIME:
														pst.setTime(parameterIndex, Time.valueOf(val));
														break;
													case Types.DATE:
														pst.setDate(parameterIndex, Date.valueOf(val));
														break;
													case Types.TIMESTAMP:
														pst.setTimestamp(parameterIndex, Timestamp.valueOf(val));
														break;
													case Types.DECIMAL:
														pst.setBigDecimal(parameterIndex, new BigDecimal(val));
														break;
													case Types.NUMERIC:
														pst.setBigDecimal(parameterIndex, new BigDecimal(val));
														break;
													case Types.FLOAT:
													case Types.REAL:
														pst.setFloat(parameterIndex, Float.valueOf(val));
														break;
													case Types.DOUBLE:
														pst.setDouble(parameterIndex, Double.valueOf(val));
														break;
													case Types.BIT:
													case Types.BOOLEAN:
														pst.setBoolean(parameterIndex, Boolean.valueOf(val));
														break;
													default:
														throw new UnsupportedOperationException("Unsupported sql type: " + columnType + " java.sql.Types." + JdbcTypeConverter.getTypeName(new Integer(columnType)));
												}
											} catch (Throwable thr) {
												throw new TestExecutionError("Error loading row[" + rowNum + "], columnIndex=[" + i + "], column=[" + col + "], colType[" + columnType + "], value='" + val + "'", DatabaseConstants.ERR_DATA_TYPE_CONVERSION, thr);
											}
										}
									}

									// add as a batch
									pst.addBatch();

									if (rowNum >= 5000) {
										pst.executeBatch();
										rowNum = 0;
									}
								}

								if (rowNum != 0) {
									pst.executeBatch();
								}

								long endTime = System.currentTimeMillis();
								applicationLog.info("Inserted [" + rowsInserted + "] rows in [" + (endTime - startTime) + "] ms.");
							} finally {
								fdata.dispose();
							}
						}
					});
				} catch (TestExecutionError e) {
					throw e;
				} catch (Exception e) {
					throw new TestExecutionError("", e);
				}
			}
		}
	}

	static DatabaseImplementation getDatabaseImplementation(DatabaseFeatureModule parent, DatabaseConnection conn) throws TestExecutionError {
		DatabaseImplementation impl = parent.getImplementation(conn.getId());

		if (impl == null) {
			throw new TestExecutionError("Database implementation id " + conn.getId() + " doesn't exist", DatabaseConstants.ERR_MISSING_DATABASE_IMPLEMENTATION);
		}
		return impl;
	}

	static void validateMode(DatabaseConnection conn, String mode, VariableContext context) throws TestExecutionError {
		// validate that this connection was prepared for this test
		if (!context.hasVariableBeenDeclared("database")) {
			throw new TestExecutionError("Database support not available for this test", DatabaseConstants.ERR_MISSING_DATABASE_SUPPORT);
		}

		ETLTestValueObject database = context.getValue("database");

		ETLTestValueObject connQuery = database.query("connections." + conn.getId() + "." + (mode == null ? "default" : mode));

		if (connQuery == null) {
			throw new TestExecutionError("Database not prepared for connection id " + conn.getId() + ", mode " + mode, DatabaseConstants.ERR_DATABASE_MODE_NOT_PREPARED);
		}
	}

	static String getDefaultMode(DatabaseConnection conn, ETLTestMethod method) throws TestExecutionError {
		String mode = null;

		// validate the mode
		// the search order for modes is like this:
		// 1 - Look at method annotations.  If this connection is present, use the first mode declared
		// 2 - Look at class annotations.  If this connection is present, use the first mode declared
		List<ETLTestAnnotation> annot = new ArrayList<ETLTestAnnotation>();

		annot.addAll(method.getAnnotations("@Database"));
		annot.addAll(method.getTestClass().getAnnotations("@Database"));

		for (ETLTestAnnotation ann : annot) {
			ETLTestValueObject annValue = ann.getValue();
			if (annValue.query("id").getValueAsString().equals(conn.getId())) {
				ETLTestValueObject modes = annValue.query("modes");
				if (modes != null) {
					mode = modes.getValueAsList().get(0).getValueAsString();

					break;
				}
			}
		}

		return mode;
	}

	@Override
	public action_code _assert(AssertRequest request, ETLTestMethod mt, ETLTestOperation testOperation, ETLTestValueObject valueObject, VariableContext variableContext, ExecutionContext executionContext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		return assertHandler._assert(request, mt, testOperation, valueObject, variableContext, executionContext);
	}

	@Override
	public action_code executeSql(ExecuteSqlRequest request, ETLTestMethod mt, ETLTestOperation testOperation, ETLTestValueObject valueObject, VariableContext variableContext, ExecutionContext executionContext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		return executeHandler.executeSql(request, mt, testOperation, valueObject, variableContext, executionContext);
	}

	@Override
	public action_code executeSqlProcedure(ExecuteSqlProcedureRequest request, ETLTestMethod mt, ETLTestOperation testOperation, ETLTestValueObject valueObject, VariableContext variableContext, ExecutionContext executionContext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		return executeHandler.executeSqlProcedure(request, mt, testOperation, valueObject, variableContext, executionContext);
	}

	@Override
	public action_code executeSqlScript(ExecuteSqlScriptRequest request, ETLTestMethod mt, ETLTestOperation testOperation, ETLTestValueObject valueObject, VariableContext variableContext, ExecutionContext executionContext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		return executeHandler.executeSqlScript(request, mt, testOperation, valueObject, variableContext, executionContext);
	}

	static class ConnectionMode {
		String connectionId;
		DatabaseConnection conn;
		List<String> modes = new ArrayList<String>();

		String getMode() {
			if (modes != null) {
				return modes.get(0);
			}

			return null;
		}

		String getPrettyMode() {
			String mode = getMode();

			return mode != null ? mode : "default";
		}
	}

	static List<ConnectionMode> getConnection(DatabaseFeatureModule parent, ETLTestValueObject etlTestValueObject, VariableContext context, ETLTestMethod method) throws TestExecutionError {
		List<ConnectionMode> connModes = new ArrayList<ConnectionMode>();

		DatabaseConfiguration databaseConfiguration = parent.getDatabaseConfiguration();

		// query for 'connections'
		ETLTestValueObject connectionsObj = etlTestValueObject.query("connections");

		if (connectionsObj != null) {
			Map<String, ETLTestValueObject> valueAsMap = connectionsObj.getValueAsMap();
			for (Map.Entry<String, ETLTestValueObject> connectionObj : valueAsMap.entrySet()) {
				ConnectionMode connectionMode = new ConnectionMode();

				connectionMode.connectionId = connectionObj.getKey();

				// we either have a 'mode' here, or 'modes'
				readModes(connectionMode, connectionObj.getValue());

				connModes.add(connectionMode);
			}
		} else {
			ConnectionMode connMode = new ConnectionMode();
			connModes.add(connMode);

			// determine the connection id and mode
			ETLTestValueObject connectionIdObj = etlTestValueObject.query("connection-id");

			final DatabaseConnection conn;

			if (connectionIdObj != null) {
				connMode.connectionId = connectionIdObj.getValueAsString();
			} else {
				// go through the search order:
				// 1 - Annotation on the method
				// 2 - Annotation on the class
				// 3 - The default connection
				List<ETLTestAnnotation> annot = method.getAnnotations("@Database");

				if (annot.size() != 0) {
					// this is the one - use the first
					connMode.connectionId = annot.get(0).getValue().query("id").getValueAsString();
				} else {
					// move on to the class annotations
					annot = method.getTestClass().getAnnotations("@Database");

					if (annot.size() != 0) {
						// this is the one - use the first
						connMode.connectionId = annot.get(0).getValue().query("id").getValueAsString();
					} else {
						// move on to the default
						// bail
						if (!databaseConfiguration.hasDefaultConnection()) {
							throw new TestExecutionError("Connection not specified and no default has been set", DatabaseConstants.ERR_NO_CONNECTION_SPECIFIED);
						}

						connMode.connectionId = databaseConfiguration.getDefaultConnection().getId();
					}
				}
			}

			// now check for the mode(s)
			readModes(connMode, etlTestValueObject);
		}

		// now that that's done, look up the info in the configuration for validation and set the default mode if not specified
		Iterator<ConnectionMode> it = connModes.iterator();

		while (it.hasNext()) {
			ConnectionMode cMode = it.next();

			cMode.conn = databaseConfiguration.getDatabaseConnection(cMode.connectionId);

			if (cMode.conn == null) {
				throw new TestExecutionError("Connection " + cMode.conn + " is not valid", DatabaseConstants.ERR_INVALID_DATABASE_CONNECTION);
			}

			if (cMode.modes.size() == 0) {
				// resolve the modes for the connection
				cMode.modes.add(getDefaultMode(cMode.conn, method));
			}

			for (String mode : cMode.modes) {
				validateMode(cMode.conn, mode, context);
			}
		}

		return connModes;
	}

	private static void readModes(ConnectionMode connMode, ETLTestValueObject etlTestValueObject) throws TestExecutionError {
		// look for multiple mode and modes attributes
		if (etlTestValueObject.query("mode") != null && etlTestValueObject.query("modes") != null) {
			throw new TestExecutionError("Attributes mode and modes are exclusive");
		}

		ETLTestValueObject modeObj = etlTestValueObject.query("mode");

		if (modeObj != null) {
			connMode.modes.add(modeObj.getValueAsString());
		} else {
			ETLTestValueObject modesObj = etlTestValueObject.query("modes");

			if (modesObj != null) {
				for (String str : modesObj.getValueAsListOfStrings()) {
					connMode.modes.add(str);
				}
			}
		}
	}

	public Feature getFeature() {
		return parent;
	}

	public action_code stage(StageRequest operation, ETLTestMethod mt, ETLTestOperation op, ETLTestValueObject obj, VariableContext context, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		stage(mt, op, obj, operation, context, econtext);

		return action_code.handled;
	}

	public action_code truncate(TruncateRequest operation, ETLTestMethod mt, ETLTestOperation op, ETLTestValueObject obj, VariableContext context, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		final String target = operation.getTarget();

		final ConnectionMode connMode = getConnection(parent, obj, context, op.getTestMethod()).get(0);

		DatabaseConnection conn = connMode.conn;

		String mode = connMode.getMode();

		DatabaseImplementation impl = getDatabaseImplementation(parent, conn);

		final String schema = operation.getSchema() == null ? impl.getDefaultSchema(conn, mode) : operation.getSchema();

		// get the connection and blast the sucker
		jdbcClient.useStatement(conn, mode, new JDBCClient.StatementClient() {
			public void connection(Connection conn, Statement st, DatabaseConnection connection, String mode, int id) throws Exception {
				String sql = "DELETE FROM " + (schema != null ? (schema + ".") : "") + target;
				applicationLog.info("Using truncate sql [" + connection.getId() + "." + connMode.getPrettyMode() + "]:\n" + sql);
				st.executeUpdate(sql);
			}
		});

		return action_code.handled;
	}

	class MigrateInfo {
		String sourceValue;
		String targetValue;
	}

	class MigrateInfos {
		MigrateInfo tableInfo;
		MigrateInfo connectionInfo;
		MigrateInfo schemaInfo;
		MigrateInfo modeInfo;
	}

	public action_code migrate(MigrateRequest operation, ETLTestMethod mt, ETLTestOperation op, ETLTestValueObject obj, VariableContext context, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		MigrateInfos mi = readMigrateExclusiveOptions(operation);

		// this process is three steps:
		// 1 - run truncate on the target
		// 2 - run extract on the source
		// 3 - run stage on the target
		ETLTestValueObject truncateValueObj = obj.copy();

		/* truncate has to see:
		-- target    << target-table
		-- optionally:
		--  connection-id << passes through
		--  schema
		--  mode
		-- for this purpose, we will need to convert the target-* properties to
		-- the unqualified properties truncate wants to see.
		*/

		ETLTestValueObjectBuilder builder = new ETLTestValueObjectBuilder(truncateValueObj);

		// remove all table attributes and set
		builder = builder.removeAllKeys("source-table", "target-table", "table");
		builder = builder.key("target").value(mi.tableInfo.targetValue);

		// remove all connection attributes and set
		builder = builder.removeAllKeys("source-connection", "target-connection", "connection");
		if (mi.connectionInfo.targetValue != null) {
			builder = builder.key("connection-id").value(mi.connectionInfo.targetValue);
		}

		// remove all schema attributes and set
		builder = builder.removeAllKeys("source-schema", "target-schema", "schema");
		if (mi.schemaInfo.targetValue != null) {
			builder = builder.key("schema").value(mi.schemaInfo.targetValue);
		}

		// remove all mode attributes and set
		builder = builder.removeAllKeys("source-mode", "target-mode", "mode");
		if (mi.modeInfo.targetValue != null) {
			builder = builder.key("mode").value(mi.modeInfo.targetValue);
		}

		// create the operation and dispatch
		ETLTestOperation truncateOp = op.createSibling("truncate", truncateValueObj);

		econtext.process(truncateOp, context);

		/* extract needs:
		-- source-table *
		-- connection-id *
		-- mode *
		-- target-file *
		-- source-schema *
		-- column-list *
		-- column-list-mode *
		 */

		ETLTestValueObject extractValueObj = obj.copy();
		builder = new ETLTestValueObjectBuilder(extractValueObj);

		// remove all table attributes and set
		builder = builder.removeAllKeys("source-table", "target-table", "table");
		builder = builder.key("source-table").value(mi.tableInfo.sourceValue);

		// remove all connection attributes and set
		builder = builder.removeAllKeys("source-connection", "target-connection", "connection");
		if (mi.connectionInfo.sourceValue != null) {
			builder = builder.key("connection-id").value(mi.connectionInfo.sourceValue);
		}

		// remove all schema attributes and set
		builder = builder.removeAllKeys("source-schema", "target-schema", "schema");
		if (mi.schemaInfo.sourceValue != null) {
			builder = builder.key("source-schema").value(mi.schemaInfo.sourceValue);
		}

		// remove all mode attributes and set
		builder = builder.removeAllKeys("source-mode", "target-mode", "mode");
		if (mi.modeInfo.sourceValue != null) {
			builder = builder.key("mode").value(mi.modeInfo.sourceValue);
		}

		ConnectionMode connMode = getConnection(parent, extractValueObj, context, op.getTestMethod()).get(0);

		DatabaseConnection conn = connMode.conn;

		File targetFile = databaseRuntimeSupport.getGeneratedDataSet(conn, connMode.getPrettyMode(), op.getQualifiedName() + "_extract");

		// remove all mode attributes and set
		builder = builder.key("target-file").value(targetFile.getAbsolutePath());

		// create the operation and dispatch
		ETLTestOperation extractOp = op.createSibling("extract", extractValueObj);

		econtext.process(extractOp, context);

		/* stage needs:
		-- source-file *
		-- target-table *
		-- connection-id *
		-- target-schema *
		-- mode *
		 */
		ETLTestValueObject stageValueObj = obj.copy();
		builder = new ETLTestValueObjectBuilder(stageValueObj);

		// remove all table attributes and set
		builder = builder.removeAllKeys("source-table", "target-table", "table");
		builder = builder.key("source-file").value(targetFile.getAbsolutePath());
		builder = builder.key("target-table").value(mi.tableInfo.targetValue);

		// remove all connection attributes and set
		builder = builder.removeAllKeys("source-connection", "target-connection", "connection");
		if (mi.connectionInfo.targetValue != null) {
			builder = builder.key("connection-id").value(mi.connectionInfo.targetValue);
		}

		// remove all schema attributes and set
		builder = builder.removeAllKeys("source-schema", "target-schema", "schema");
		if (mi.schemaInfo.targetValue != null) {
			builder = builder.key("source-schema").value(mi.schemaInfo.targetValue);
		}

		// remove all mode attributes and set
		builder = builder.removeAllKeys("source-mode", "target-mode", "mode");
		if (mi.modeInfo.targetValue != null) {
			builder = builder.key("mode").value(mi.modeInfo.targetValue);
		}

		// create the operation and dispatch
		ETLTestOperation stageOp = op.createSibling("stage", stageValueObj);

		econtext.process(stageOp, context);

		return action_code.handled;
	}

	private MigrateInfos readMigrateExclusiveOptions(MigrateRequest operation) throws TestExecutionError {
		// table
		String table = operation.getTable();
		String sourceTable = operation.getSourceTable();
		String targetTable = operation.getTargetTable();

		// connection ID
		String conId = operation.getConnectionId();
		String sourceConId = operation.getSourceConnectionId();
		String targetConId = operation.getTargetConnectionId();

		// schema
		String schema = operation.getSchema();
		String sourceSchema = operation.getSourceSchema();
		String targetSchema = operation.getTargetSchema();

		// mode
		String mode = operation.getMode();
		String sourceMode = operation.getSourceMode();
		String targetMode = operation.getTargetMode();

		MigrateInfos mi = new MigrateInfos();

		mi.tableInfo = checkExclusive(table, sourceTable, targetTable, "table");
		mi.connectionInfo = checkExclusive(conId, sourceConId, targetConId, "connection-id");
		mi.schemaInfo = checkExclusive(schema, sourceSchema, targetSchema, "schema");
		mi.modeInfo = checkExclusive(mode, sourceMode, targetMode, "mode");

		/* there are a few error conditions here:
		-- 1 - If the source and target tables are the same, meaning
		       that the table, connection, schema and mode are all the same
		*/

		if (
				compare(mi.tableInfo.sourceValue, mi.tableInfo.targetValue)
						&&
						compare(mi.connectionInfo.sourceValue, mi.connectionInfo.targetValue)
						&&
						compare(mi.schemaInfo.sourceValue, mi.schemaInfo.targetValue)
						&&
						compare(mi.modeInfo.sourceValue, mi.modeInfo.targetValue)) {
			throw new TestExecutionError("Source and target tables are the same", DatabaseConstants.ERR_SOURCE_TARGET_IDENTICAL);
		}

		return mi;
	}

	private boolean compare(String s, String t) {
		if (s == null && t == null) {
			return true;
		} else if (s == null && t != null) {
			return false;
		} else if (s != null && t == null) {
			return false;
		} else if (s == t) {
			return true;
		} else if (s.equals(t)) {
			return true;
		}

		return false;
	}

	private MigrateInfo checkExclusive(String conId, String sourceConId, String targetConId, String baseAttribute) throws TestExecutionError {
		MigrateInfo mt = new MigrateInfo();

		if (conId != null || sourceConId != null || targetConId != null) {
			// if the generic is specified and either of the specific ones is specified, fail
			if (conId != null && (sourceConId != null || targetConId != null)) {
				throw new TestExecutionError("When " + baseAttribute + " is specified, neither of source-" + baseAttribute + " or target-" + baseAttribute + " may be specified", DatabaseConstants.ERR_GENERIC_AND_SPECIFIC_OPTIONS_FAIL);
			}

			// if the generic is NOT specified, then both of the specific operands are required
			if (conId == null && (sourceConId == null || targetConId == null)) {
				throw new TestExecutionError("When " + baseAttribute + " is not specified, both source-" + baseAttribute + " and target-" + baseAttribute + " must be specified", DatabaseConstants.ERR_GENERIC_AND_SPECIFIC_OPTIONS_FAIL);
			}

			mt.sourceValue = conId == null ? sourceConId : conId;
			mt.targetValue = conId == null ? targetConId : conId;
		}

		return mt;
	}

	@Override
	public void beginPackage(String name, VariableContext context) throws TestAssertionFailure, TestExecutionError, TestWarning {
		// record the meta data files that exist in this test
		MetaDataContext dataMetaContext = parent.getDataMetaContext();
		MetaDataContext databaseMetaContext = parent.getDatabaseMetaContext();

		MetaDataPackageContext dataPackageContext = dataMetaContext.createPackageContextForCurrentTest(MetaDataPackageContext.path_type.test_source);
		try {
			MetaDataArtifact dataPackageContextArtifact = dataPackageContext.createArtifact("data", databaseRuntimeSupport.getSourceDataDirectoryForCurrentTest());
			dataPackageContextArtifact.populateAllFromDir();
		} catch (IOException e) {
			throw new TestExecutionError("Error creating meta data context", e);
		}

		dataPackageContext = databaseMetaContext.createPackageContextForCurrentTest(MetaDataPackageContext.path_type.test_source);
		try {
			MetaDataArtifact sqlPackageContextArtifact = dataPackageContext.createArtifact("sql", databaseRuntimeSupport.getSourceScriptsDirectoryForCurrentTest());
			sqlPackageContextArtifact.populateAllFromDir();
		} catch (IOException e) {
			throw new TestExecutionError("Error creating meta data context", e);
		}
	}
}